#include <stdlib.h>
#include "RedBlack.h"

// ---------------------------------------------------------------- Definitions

   // Return the minumum of two numbers
inline static int
MIN(int a, int b) {
   return  (a < b) ? a : b;
}

   // Return the maximum of two numbers
inline static int
MAX(int a, int b) {
   return  (a > b) ? a : b;
}

// ----------------------------------------------- Constructors and Destructors

RedBlackNode::RedBlackNode(ItemType * item, color_t color)
   : myData(item), myColor(color)
{
   Reset();
}

RedBlackNode::~RedBlackNode(void) {
   if (mySubtree[LEFT])  delete  mySubtree[LEFT];
   if (mySubtree[RIGHT]) delete  mySubtree[RIGHT];
}

// ------------------------------------------------- Rotating and Re-Balancing

void
RedBlackNode::ColorFlip() {
      // Determine the opposite color of this node
   color_t  otherColor = Opposite(myColor);
      // Make subtrees the same color as this node
   for (int i = 0; i < MAX_SUBTREES; i++) {
      if (mySubtree[i])  mySubtree[i]->myColor = myColor;
   }
      // Flip the color of this node
   myColor = otherColor;
}

void
RedBlackNode::RotateOnce(RedBlackNode * & root, dir_t  dir)
{
   dir_t  otherDir = Opposite(dir);
   RedBlackNode * oldRoot = root;

      // assign new root
   root = oldRoot->mySubtree[otherDir];

      // new-root exchanges it's "dir" mySubtree for it's parent
   oldRoot->mySubtree[otherDir] = root->mySubtree[dir];
   root->mySubtree[dir] = oldRoot;
}

void
RedBlackNode::RotateTwice(RedBlackNode * & root, dir_t  dir)
{
   dir_t  otherDir = Opposite(dir);
   RedBlackNode * oldRoot = root;
   RedBlackNode * oldOtherDirSubtree = root->mySubtree[otherDir];

      // assign new root
   root = oldRoot->mySubtree[otherDir]->mySubtree[dir];

      // new-root exchanges it's "dir" mySubtree for it's grandparent
   oldRoot->mySubtree[otherDir] = root->mySubtree[dir];
   root->mySubtree[dir] = oldRoot;

      // new-root exchanges it's "other-dir" mySubtree for it's parent
   oldOtherDirSubtree->mySubtree[dir] = root->mySubtree[otherDir];
   root->mySubtree[otherDir] = oldOtherDirSubtree;
}

void
RedBlackNode::FixRedViolation(RedBlackNode * & parent,
                              RedBlackNode * & child,
                              dir_t  dir)
{
      // See if this is the top of the tree
   if (parent == NULL) {
         // Yup - this node is the root so just
         // make it black and we are done.
      child->myColor = BLACK;
   } else {
         // Figure out which subtree of the parent I am
      dir_t  thisDir  = (parent->mySubtree[LEFT] == child) ? LEFT : RIGHT;
         // Get the direction of my sibling
      dir_t  otherDir = Opposite(thisDir);
         // Get my sibling and his color
      RedBlackNode * sibling = parent->mySubtree[otherDir];
      color_t  sibColor = (sibling) ? sibling->myColor : BLACK;
      if (sibColor == BLACK) {
            // Need to perform a rotation
         if (thisDir == dir) {
               // Need a single rotation
            RotateOnce(parent, otherDir);
         } else {
               // Need a double rotation
            RotateTwice(parent, otherDir);
         }
      }
         // Now color flip the parent
      parent->ColorFlip();
   }
}

int
RedBlackNode::FixBlackViolation(RedBlackNode * & root, dir_t  dir)
{

#define  CASE(x)   /*  cout << "*** case " << x << " ***\n"  */

   int  propagated = 0;   // Initialize the return value

      // "dir" is the direction that just became deficient, the other
      // direction is the side that wasnt deleted from.
   dir_t  otherDir = Opposite(dir);

      // Get the deficient subtree (if its exists) and its sibling
   RedBlackNode * child   = root->mySubtree[dir];
   RedBlackNode * sibling = root->mySubtree[otherDir];

   if (Color(child) == RED) {
      CASE(0);
         // Increase the black height of this entire subtree by
         // simply making the child black. In theory - this routine
         // will never be called if the child is red.
      child->myColor = BLACK;
   } else if (Color(sibling) == BLACK) {
         // Need to know if the parent is red or not. If it is, then
         // any rotated result needs to have a red-root (and root will
         // need to change to black), otherwise any rotated result will
         // have a black root.
      int  blackParent = (Color(root) == BLACK);
      if (! blackParent)  root->myColor = BLACK;
         // Now look at the color of the sibling's children
      if (Color(sibling->mySubtree[dir]) == RED) {
         CASE((blackParent ? 1 : 4));
            // Need to change the color of this subtree if the parent is black
         if (blackParent)  sibling->mySubtree[dir]->myColor = BLACK;
            // Now rotate twice to make this subtree the new root
         RotateTwice(root, dir);
      } else if (Color(sibling->mySubtree[otherDir]) == RED) {
         CASE((blackParent ? 2 : 5));
            // Need to change the color of this subtree
         sibling->mySubtree[otherDir]->myColor = BLACK;
            // If the root is red then exchange its color with the sibling
         if (! blackParent)  sibling->myColor = RED;
            // Now rotate once to make the sibling the new root
         RotateOnce(root, dir);
      } else {
         CASE((blackParent ? 3 : 6));
            // Make the sibling red
         sibling->myColor = RED;
            // If the root was black we have fixed the black-property
            // violation at this level by shortening the black height of
            // the "dir" subtree, hence the problem has been propagated
            // up to the next higher level.
         propagated = blackParent;
      }
   } else {   // Color(sibling) == RED
         // First, perform a single rotation and flip the color of the
         // sibling and its "dir" subtree.
      sibling->myColor = BLACK;
      sibling->mySubtree[dir]->myColor = RED;
      RotateOnce(root, dir);
         // Now check for a red-property violation with the sibling's
         // subtree (which after the rotation, is now in a different
         // position). Note that the sibling's former subtree is the
         // "nephew" or "niece" of the child.
      RedBlackNode * & niece = root->mySubtree[dir]->mySubtree[otherDir];
      if ((Color(niece->mySubtree[dir]) == RED)  &&
          (Color(niece->mySubtree[otherDir]) == RED)) {
            // change the color of the niece
         niece->myColor = BLACK;
            // See if child doesnt exist
            // (this time - a double)
         if (child) {
            CASE("7a");
         } else {
            CASE("7b");
               // No child, need an additional rotation
               // (this time a double)
            RotateTwice(root->mySubtree[dir], dir);
         }
      } else if (Color(niece->mySubtree[dir]) == RED) {
         CASE(8);
         FixRedViolation(root->mySubtree[dir], niece, dir);
      } else if (Color(niece->mySubtree[otherDir]) == RED) {
         CASE(9);
         FixRedViolation(root->mySubtree[dir], niece, otherDir);
      } else {
            // No extra adjustments are necessary
         CASE(10);
      }
   }
   return  propagated;
}

// ------------------------------------------------------- Comparisons

cmp_t
RedBlackNode::Compare(KeyType key, cmp_t cmp) const
{
   switch (cmp) {
      case EQ_CMP :  // Standard comparison
         return  myData->Compare(key);

      case MIN_CMP :  // Find the minimal element in this tree
         return  (mySubtree[LEFT] == NULL) ? EQ_CMP : MIN_CMP;

      case MAX_CMP :  // Find the maximal element in this tree
         return  (mySubtree[RIGHT] == NULL) ? EQ_CMP : MAX_CMP;
   }
}

// ------------------------------------------------------- Search/Insert/Delete

ItemType *
RedBlackNode::SearchRB(KeyType key, RedBlackNode * root, cmp_t cmp)
{
   cmp_t result;
   while (root  &&  (result = root->Compare(key, cmp))) {
      root = root->mySubtree[(result < 0) ? LEFT : RIGHT];
   }
   return  (root) ? root->myData : NULL;
}

ItemType *
RedBlackNode::InsertRB(ItemType * item, RedBlackNode * & root) {
   RedBlackNode * parent = NULL;
   return  InsertRB(item, root, parent);
}

ItemType *
RedBlackNode::DeleteRB(KeyType key, RedBlackNode * & root, cmp_t cmp) {
   int blackDecrease = 0;
   return  DeleteRB(key, root, blackDecrease, cmp);
      // NOTE: If blackDecrease is true at the root, then the tree
      //       has been fixed so that both sides "shortened" by the
      //       same amount - so no corrective action needs to be taken.
}

ItemType *
RedBlackNode::InsertRB(ItemType * item, RedBlackNode * & root,
                                        RedBlackNode * & parent)
{
      // See if the tree is empty
   if (root == NULL) {
         // Insert new node here. If parent is null this is
         // the first node so it should be black - otherwise
         // make it red.
      root = new RedBlackNode(item, (parent == NULL) ? BLACK : RED);
      return  NULL;
   }

      // Compare items and determine which direction to search
   ItemType * found = NULL;
   cmp_t  result = root->Compare(item->Key());
   dir_t  dir = (result == MIN_CMP) ? LEFT : RIGHT;

   if (result == EQ_CMP) {
         // Key already in tree at this node
      return  root->myData;
   }
      // Insert into "dir" subtree 
   found = InsertRB(item, root->mySubtree[dir], root);
   if (found)  return  found;   // already here - dont insert

      // Check for red-property violation with child
   if ((root->myColor == RED) &&
       (root->mySubtree[dir]->myColor == RED)) {
      FixRedViolation(parent, root, dir);
   }

   return  found;
}

ItemType *
RedBlackNode::DeleteRB(KeyType key, RedBlackNode * & root,
                                    int & blackDecrease, cmp_t cmp)
{
      // Initialize
   ItemType * found = NULL;
   blackDecrease = 0;

      // See if the tree is empty
   if (root == NULL) {
         // Key not found
      return  NULL;
   }

      // Compare items and determine which direction to search
   cmp_t    result = root->Compare(key, cmp);
   dir_t    dir = (result == MIN_CMP) ? LEFT : RIGHT;

   if (result != EQ_CMP) {
         // Delete from "dir" subtree 
      found = DeleteRB(key, root->mySubtree[dir], blackDecrease, cmp);
      if (! found)  return  found;   // not found - can't delete
   } else  {   // Found key at this node
      found = root->myData;  // set return value

      if ((root->mySubtree[LEFT] == NULL) &&
          (root->mySubtree[RIGHT] == NULL)) {
             // We have a leaf -- remove it
         blackDecrease = (root->myColor == BLACK);
         delete  root;
         root = NULL;
         return  found;
      } else {
            // If we have a left subtree, replace ourself with our predecessor,
            // otherwise replace ourself with our successor. 
         dir = (root->mySubtree[LEFT]) ? LEFT : RIGHT;
         cmp = (dir == LEFT) ? MAX_CMP : MIN_CMP;
         root->myData = DeleteRB(key, root->mySubtree[dir],
                                      blackDecrease, cmp);
      }
   }

      // Look for black-property violations
   if (blackDecrease) {
      blackDecrease = FixBlackViolation(root, dir);
   }

   return  found;
}

// --------------------------------------------------------------- Verification

int
RedBlackNode::Height() const {
   int  leftHeight  = (mySubtree[LEFT])  ? mySubtree[LEFT]->Height()  : 0;
   int  rightHeight = (mySubtree[RIGHT]) ? mySubtree[RIGHT]->Height() : 0;
      // If this node is red, then it doesnt add to the logical height
      // of the tree (this is why RED is defined to be zero)
   return  (myColor + MAX(leftHeight, rightHeight));
}

int
RedBlackNode::CheckRB() const {
   int  valid = 1;

      // First verify that subtrees are correct
   if (mySubtree[LEFT])   valid *= mySubtree[LEFT]->CheckRB();
   if (mySubtree[RIGHT])  valid *= mySubtree[RIGHT]->CheckRB();

      // Verify that Black property is satisfied
   int  leftHeight  = (mySubtree[LEFT])  ? mySubtree[LEFT]->Height()  : 0;
   int  rightHeight = (mySubtree[RIGHT]) ? mySubtree[RIGHT]->Height() : 0;
   if (leftHeight != rightHeight) {
      valid = 0;
      cerr << "Node " << Key() << " has subtrees of unequal height" << endl;
   }

      // Verify that Red property is satisfied
   if (myColor == RED) {
      for (int i = 0; i < MAX_SUBTREES; ++i) {
         if (mySubtree[i]  &&  (mySubtree[i]->myColor == RED)) {
            valid = 0;
            cerr << "Red node " << Key() << " has a RED subtree"
                 << "[" << i << "]" << endl;
         }
      }
   }

      // Verify that search-tree property is satisfied
   if ((mySubtree[LEFT]) && (mySubtree[LEFT]->Compare(Key()) == MIN_CMP)) {
      valid = 0;
      cerr << "Node " << Key() << " is *smaller* than left subtree"
           << mySubtree[LEFT]->Key() << endl;
   }
   if ((mySubtree[RIGHT]) && (mySubtree[RIGHT]->Compare(Key()) == MAX_CMP)) {
      valid = 0;
      cerr << "Node " << Key() << " is *greater* than right subtree"
           << mySubtree[RIGHT]->Key() << endl;
   }

   return  valid;
}

//----------------------------------------------- Routines for dumping the tree

static inline ostream &
Indent(ostream & os, int len) {
   for (int i = 0; i < len; i++) {
      os << ' ';
   }
   return  os;
}

enum TraversalOrder { LTREE, KEY, RTREE };

static void
DumpRB(ostream & os, TraversalOrder order, const RedBlackNode * node,
                                           int level=0)
{
    unsigned  len = (level * 5) + 1;
    if ((order == LTREE) && (node->Subtree(RedBlackNode::LEFT) == NULL)) {
       Indent(os, len) << "     **NULL**" << endl;
    }
    if (order == KEY) {
       char  colorChar = (node->Color() == RedBlackNode::RED) ? 'R' : 'B';
       Indent(os, len) << node->Key() << ":" << colorChar << endl;
    }
    if ((order == RTREE) && (node->Subtree(RedBlackNode::RIGHT) == NULL)) {
       Indent(os, len) << "     **NULL**" << endl;
    }
}

static void
DumpRB(ostream & os, const RedBlackNode * node, int level=0)
{
   if (node == NULL) {
      os << "***EMPTY TREE***" << endl;
   } else {
      DumpRB(os, RTREE, node, level);
      if (node->Subtree(RedBlackNode::RIGHT)  !=  NULL) {
         DumpRB(os, node->Subtree(RedBlackNode::RIGHT), level+1);
      }
      DumpRB(os, KEY, node, level);
      if (node->Subtree(RedBlackNode::LEFT)  !=  NULL) {
         DumpRB(os, node->Subtree(RedBlackNode::LEFT), level+1);
      }
      DumpRB(os, LTREE, node, level);
   }// if non-empty tree
}

void
RedBlackTree::DumpTree(ostream & os) const {
   DumpRB(os, myRoot);
}

const char RedBlackTree::ourKind[] = "Red-Black tree";
