#ifndef REDBLACK_H
#define REDBLACK_H

#include <stddef.h>
#include "Item.h"

class  ostream;

// RedBlackNode -- Class to implement a Red-Black Tree (bottom up)
//
class RedBlackNode {
public:
      // Max number of subtrees per node
   enum  { MAX_SUBTREES = 2 };

      // Indices into a subtree array
   enum  dir_t { LEFT = 0, RIGHT = 1 };

      // Node colors
   enum color_t { RED = 0, BLACK = 1 };

      // Return the opposite of the given color
   static color_t
   Opposite(color_t  color) {
     return color_t(1 - int(color));
   }

      // Return the opposite direction of the given index
   static  dir_t
   Opposite(dir_t  dir) { 
     return dir_t(1 - int(dir));
   }

   // ----- Constructors and destructors: 

   RedBlackNode(ItemType * item=NULL, color_t color=RED);
   virtual ~RedBlackNode(void);

   // ----- Query attributes:

      // Get this node's data
   ItemType *
   Data() const { return  myData; }

      // Get this node's key field
   KeyType
   Key() const { return  myData->Key(); }

      // Get this node's color
   color_t
   Color(void) const { return  myColor; }

      // Same as above but this works for a NULL node (we assume the
      // color of a null (external) node is black
   static color_t
   Color(RedBlackNode * tree) {
      return  (tree == NULL) ? BLACK : tree->myColor;
   }

      // Get the item at the top of the left/right subtree of this
      // item (the result may be NULL if there is no such item).
      //
   RedBlackNode *
   Subtree(dir_t dir) const { return  mySubtree[dir]; }

   // ----- Search/Insert/Delete
   //
   //   NOTE: These are all static functions instead of member functions
   //         because most of them need to modify the given tree root
   //         pointer. If these were instance member functions than
   //         that would correspond to having to modify the 'this'
   //         pointer, which is not allowed in C++. Most of the 
   //         functions that are static and which take an AVL tree
   //         pointer as a parameter are static for this reason.
   
      // Look for the given key, return NULL if not found,
      // otherwise return the item's address.
   static ItemType *
   SearchRB(KeyType key, RedBlackNode * root, cmp_t cmp=EQ_CMP);

      // Search is an alias for SearchRB
   static ItemType *
   Search(KeyType key, RedBlackNode * root, cmp_t cmp=EQ_CMP) {
      return  SearchRB(key, root, cmp);
   }

      // Insert the given key, return NULL if it was inserted,
      // otherwise return the existing item with the same key.
   static ItemType *
   InsertRB(ItemType * item, RedBlackNode * & root);

      // Insert is an alias for InsertRB
   static ItemType *
   Insert(ItemType * item, RedBlackNode * root) {
      return  InsertRB(item, root);
   }

      // Delete the given key from the tree. Return the corresponding
      // node, or return NULL if it was not found.
   static ItemType *
   DeleteRB(KeyType key, RedBlackNode * & root, cmp_t cmp=EQ_CMP);

      // Delete is an alias for DeleteRB
   static ItemType *
   Delete(KeyType key, RedBlackNode * root, cmp_t cmp=EQ_CMP) {
      return  DeleteRB(key, root, cmp);
   }

   // Verification 

      // Return the black-height of this tree (the maximum number of
      // black nodes from the root to a leaf).
   int
   Height() const;

      // Verify this tree is a valid Red-Black tree, return TRUE if it is,
      // return FALSE otherwise
   int
   CheckRB() const;

      // Check is an alias for CheckRB
   int
   Check() const {
      return  CheckRB();
   }

      // If you want to provide your own allocation scheme than simply
      // #define the preprocessor manifest constant named CUSTOM_ALLOCATE
      // and make sure you provide and link with your own overloaded
      // versions of operators "new" and "delete" for this class.
#ifdef CUSTOM_ALLOCATE
   void *
   operator  new(size_t);

   void
   operator  delete(void *);
#endif  /* CUSTOM_ALLOCATE */


private:

   // ----- Private data

   ItemType * myData;    // Data field
   RedBlackNode  * mySubtree[MAX_SUBTREES];   // Pointers to subtrees
   color_t    myColor;   // Node color

      // Reset all subtrees to null
   void
   Reset(void) {
      mySubtree[LEFT] = mySubtree[RIGHT] = NULL ;
   }

   // ----- Routines that do the *real* insertion/deletion

      // Insert the given key into the given tree. Return the node if
      // it already exists. Otherwise return NULL to indicate that the
      // key was successfully inserted. If parent is NULL, then root
      // is the root for the entire tree, otherwise root is an immediate
      // subtree of parent.
   static ItemType *
   InsertRB(ItemType * item, RedBlackNode * & root,
                             RedBlackNode * & parent);

      // Delete the given key from the given tree. Return NULL if the
      // key is not found in the tree. Otherwise return a pointer to the
      // node that was removed from the tree. Upon return, "blackDecrease"
      // will indicate if a black-property violation exists due to a
      // black node being deleted.
   static ItemType *
   DeleteRB(KeyType key, RedBlackNode * & root,
                         int & blackDecrease, cmp_t cmp=EQ_CMP);

   // Routines for rebalancing and rotating subtrees

      // Make this node the given color and its children
      // the opposite color.
   void ColorFlip();

      // Perform an XX rotation for the given direction 'X'.
   static void
   RotateOnce(RedBlackNode * & root, dir_t  dir);

      // Perform an XY rotation for the given direction 'X'
   static void
   RotateTwice(RedBlackNode * & root, dir_t  dir);

      // Fix a violation of the red-property (two consecutive
      // red-nodes) between the given child and its "dir"
      // subtree.
   static void
   FixRedViolation(RedBlackNode * & parent,
                   RedBlackNode * & child,
                   dir_t  dir);

      // Fix a violation of the black-property (different numer of black
      // nodes on two external paths) between the given root and its
      // "dir" subtree. Returns 0 if the violation was fixed at this level,
      //  and returns 1 if the violation was propagated upward.
   static int
   FixBlackViolation(RedBlackNode * & root, dir_t  dir);

      // Perform a comparison of the given key against the given
      // item using the given criteria (min, max, or equivalence
      // comparison). Returns:
      //   EQ_CMP if the keys are equivalent
      //   MIN_CMP if this key is less than the item's key
      //   MAX_CMP if this key is greater than item's key
   cmp_t
   Compare(KeyType key, cmp_t cmp=EQ_CMP) const;

private:
   // Disallow copying and assignment
   RedBlackNode(const RedBlackNode &);
   RedBlackNode & operator=(const RedBlackNode &);

};

   // Class RedBlackTree is a simple container object to "house" a RedBlackNode
   // that represents the root-node of and RedBlackTree. Most of the member
   // functions simply delegate to the root RedBlackNode.
class RedBlackTree {
private:
      // Disallow copying and assingment
   RedBlackTree(const RedBlackTree &);
   RedBlackTree & operator=(const RedBlackTree &);

      // Member data
   RedBlackNode * myRoot;   // The root of the tree
   static const char  ourKind[];   // What kind of tree this is

public:
      // Constructor and destructor
   RedBlackTree() : myRoot(NULL) {};
   ~RedBlackTree() { if (myRoot)  delete myRoot; }

      // Print the tree kind
   const char *
   Kind() const { return ourKind; }

      // Dump the tree to the given output stream
   void
   DumpTree(ostream & os) const;

      // See if the tree is empty
   int
   IsEmpty() const {
      return  (myRoot == NULL);
   }

      // Search, Insert, Delete, and Check
   ItemType *
   Search(KeyType key, cmp_t cmp=EQ_CMP) {
      return  RedBlackNode::SearchRB(key, myRoot, cmp);
   }

   ItemType *
   Insert(ItemType * item) {
      return  RedBlackNode::InsertRB(item, myRoot);
   }

   ItemType *
   Delete(KeyType key, cmp_t cmp=EQ_CMP) {
      return  RedBlackNode::DeleteRB(key, myRoot, cmp);
   }

   int
   Check() const {
      return  (myRoot) ? myRoot->CheckRB() : 1;
   }
};


#endif  /* REDBLACK_H */
